// The Santa Fe Stockmarket -- Implementation of Specialist class


#import "Specialist.h"
#import "BFagent.h"
#import "collections.h"    // for the index
#import "Interval.h"
#include <math.h>
#import <stdlib.h>

@implementation Specialist

+createBegin: zone {
  Specialist *obj = [super createBegin: zone];
  
  obj->maxprice = [Interval create];
  obj->minprice = [Interval create];
  obj->eta = [Interval create];
  obj->minexcess = [Interval create];
  obj->rea = [Interval create];
  obj->reb = [Interval create];
  obj->bidfrac = [Interval create];
  obj->offerfrac = [Interval create];
  obj->volume = [Interval create];
  obj->taupdecay = [Interval create];
  obj->taupnew = [Interval create];
  
  return obj;
}

+create: zone {
  Specialist *obj = [self createBegin: zone];

  obj = [obj createEnd];

  return obj;
}

/*" One instance of this class is used to manage the trading and
  set the stock price.  It also manages the market-level parameters."*/


- setMaxPrice: (Interval *)maximumPrice
{
  [maxprice set: maximumPrice];
  [maximumPrice drop];
  return self;
}


- setMinPrice: (Interval *)minimumPrice
{
  [minprice set: minimumPrice];
  [minimumPrice drop];
  return self;
}


- setTaup: (Interval *)aTaup
{
  [taupnew set: [[[[Interval constant: -1.0] quot: aTaup] expm1] neg]];
				//pj: moved here from init method
  [taupdecay set: [[Interval constant: 1.0] diff: taupnew]];
				// moved to simplify!
  [aTaup drop];
  return self;
}



/*"The specialist can be set to type 0, 1, or 2. If this variable is
set to any other value, the model will set the Specialist to type 1
and give a warning in the terminal"*/
- setSPtype: (int)i
{
  if(i != 0 && i != 1 && i != 2 && i != 3)
    { 
      printf("The specialist type chosen is invalid. ");
      abort();
    }
  sptype = (SpecialistType)i;

  return self;
}

/*" Set the maximum number of interations to be done while looking for a market clearing price"*/
- setMaxIterations: (int)someIterations
{
  maxiterations = someIterations;
  return self;
}


- setMinExcess: (Interval *)minimumExcess
{
  [minexcess set: minimumExcess];
  [minimumExcess drop];
  return self;
}


- setETA: (Interval *)ETA
{
  [eta set: ETA];
  [ETA drop];
  return self;
}

- setREA: (Interval *)REA
{
  [rea set: REA];
  [REA drop];
  return self;
}


- setREB: (Interval *)REB
{
  [reb set: REB];
  [REB drop];
  return self;
}

//pj: init method is now unnecessary. Moved IVARS inside methods, or
//reset values when needed.
//  -init


- (Interval *)performTrading: (id) agentList Market: (id) worldForSpec
/*" This is the core method that sets a succession of trial prices and
 *  asks the agents for their bids or offer at each, generally
 *  adjusting the price towards reducing |bids - offers|.  * It gets
 *  bids and offers from the agents and * adjuss the price.  Returns
 *  the final trading price, which becomes * the next market price.
 *  Various methods are implemented, but all * have the structure:
 *  1. Set a trial price 

    2. Send each agent a -getDemandAndSlope:forPrice: message and accumulate the total
 *  number of bids and offers at that price.  

    3. [In some cases] go to  1.  

    4. Return the last trial price.  "*/
{
  int mcount;
  BOOL done;
  Interval *demand = [Interval create],
    *slope = [Interval create],
    *imbalance = [Interval create],
    *dividend = [Interval create];
  Interval * slopetotal = [Interval createWith: 0.0];
  Interval * trialprice = [Interval createWith: 0.0];
  Interval * offertotal = [Interval createWith: 0.0];  //was IVAR
  Interval * bidtotal = [Interval createWith: 0.0]; //was IVAR

  id agent;  id index;

  [volume set: [Interval constant: 0.0]];

// Save previous values
  //oldbidtotal = bidtotal;  //pj: old variables were never used anywhere
  //oldoffertotal = offertotal;
  //oldvolume = volume;

  [dividend set: [worldForSpec getDividend]];
// Main loop on {set price, get demand}
  for (mcount = 0, done = NO; mcount < maxiterations && !done; mcount++) 
    {
      // Set trial price -- various methods
      switch (sptype) 
	{
	case SP_RE:
	  // Rational expectations benchmark:  The rea and reb parameters must
	  // be calculated by hand (highly dependent on agent and dividend).
	  [trialprice set: [[rea prod: dividend] sum: reb]];
	  done = YES;		// One pass
	  break;
	  
	case SP_SLOPE:
	  if (mcount == 0)
	    [trialprice set: [worldForSpec getPrice]];
	  else 
	    {
	      // Use demand and slope information from the agent to set a new
	      // price where the market should clear if the slopes are all
	      // present and correct.  Iterate until it's close or until
	      // maxiterations is reached.
	      [imbalance set: [bidtotal diff: offertotal]];
	      if ([imbalance le: minexcess] && [imbalance ge: [minexcess neg]])
		{
		  done = YES;
		  continue;
		}
	      // Update price using demand curve slope information
	      if ([slopetotal Lne: [Interval constant: 0.0]])
		[trialprice sub: [imbalance quot: slopetotal]];
	      else
		[trialprice mul: [[Interval constant: 1.0]
				   sum: [eta prod: imbalance]]];
	    }
	  break;

	case SP_RN_START_SLOPE:	// Uses the RiskNeutral price as the starting
				// point for the search instead of the
				// last price.
	  if (mcount == 0) {
	    [trialprice set: [worldForSpec getRiskNeutral]];
	  }
	  else 
	    {
	      // Use demand and slope information from the agent to set a new
	      // price where the market should clear if the slopes are all
	      // present and correct.  Iterate until it's close or until
	      // maxiterations is reached.
	      [imbalance set: [bidtotal diff: offertotal]];
	      if ([imbalance le: minexcess] && [imbalance ge: [minexcess neg]])
		{
		  done = YES;
		  continue;
		}
	      // Update price using demand curve slope information
	      if ([slopetotal Lne: [Interval constant: 0.0]])
		[trialprice sub: [imbalance quot: slopetotal]];
	      else
		[trialprice mul: [[Interval constant: 1.0]
				   sum: [eta prod: imbalance]]];
	    }
	  break;
	
	case SP_ETA:
	  //Need to use this for ANNagent.
	  if (mcount == 0)
	    {
	      [trialprice set: [worldForSpec getPrice]];
	    }	  
	  else 
	    {
	      [trialprice set: [[worldForSpec getPrice]
				 prod: [[Interval constant: 1.0]
					 sum: [eta
						prod: [bidtotal
							diff: offertotal]]]]];
	      done = YES;	// Two passes
	    }
	  break;
	}

      // Clip trial price
      [trialprice constrainMin: minprice];
      [trialprice constrainMax: maxprice];
      
      // Get each agent's requests and sum up bids, offers, and slopes
      [bidtotal set: [Interval constant: 0.0]];
      [offertotal set: [Interval constant: 0.0]];
      [slopetotal set: [Interval constant: 0.0]];
      index = [agentList begin: [self getZone]];
      while ((agent = [index next])) 
	{
	  [slope set: [Interval constant: 0.0]];
	  [demand set: [agent getDemandAndSlope: &slope
			      forPrice: [trialprice call]]];
	  [slopetotal add: slope];
	  if ([demand Lgt: [Interval constant: 0.0]])      
	    [bidtotal add: demand];
	  else if ([demand Llt: [Interval constant: 0.0]]) 
	    [offertotal sub: demand];
	  //printf("bidtotal is %f and offertotal is %f.\n",bidtotal,offertotal);
	}
      [index drop];

      // Match up the bids and offers
      [volume set: ([bidtotal gt: offertotal] ? offertotal : bidtotal)];
      [bidfrac set: ([bidtotal gt: [Interval constant: 0.0]]
		     ? [volume quot: bidtotal] : [Interval constant: 0.0])];
      [offerfrac set: ([offertotal gt: [Interval constant: 0.0]]
		       ? [volume quot: offertotal]
		       : [Interval constant: 0.0])];
    }
  
  [demand drop];
  [slope drop];
  [imbalance drop];
  [dividend drop];
  [slopetotal drop];
  [offertotal drop];
  [bidtotal drop];
  return [trialprice ret];
}

/*"Returns the volume of trade to anybody that wants, such as the observer or output objects"*/
- (Interval *)getVolume
{
  return [[volume clone] ret];
}

-(double)getVolumeFDOX {
  return [volume getMax];
}

-(double)getVolumeFDON {
  return [volume getMin];
}


- completeTrades: agentList Market: worldForSpec
/*"Updates the agents cash and position to consummate the trades
  previously negotiated in -performTrading, with rationing if
  necessary.

 * Makes the actual trades at the last trial price (which is now the
 * market price), by adjusting the agents' holdings and cash.  The
 * actual purchase/sale my be less than that requested if rationing
 * is imposed by the specialist -- usually one of "bidfrac" and
 * "offerfrac" will be less than 1.0.
 *
 * This could easiliy be done by the agents themselves, but we let
 * the specialist do it for efficiency.
 "*/
{
  Agent * agent;
  id index;
  Interval * bfp = [Interval create],
    *ofp = [Interval create],
    *tp = [Interval create],
    *profitperunit = [Interval create];
  Interval * price = [Interval createWith: 0.0]; //pj: was IVAR


  [price set: [worldForSpec getPrice]];
  [profitperunit set: [worldForSpec getProfitPerUnit]];

// Intermediates, for speed
  [bfp set: [bidfrac prod: price]];
  [ofp set: [offerfrac prod: price]];
  [tp set: [taupnew prod: profitperunit]];

// Loop over enabled agents
  index = [agentList begin: [self getZone]];

  while ((agent = [index next])) 
    {
      // Update profit (moving average) using previous position
      [agent->profit set: [[taupdecay prod: agent->profit]
			    sum: [tp prod: agent->position]]];

      // Make the actual trades
      if ([agent->demand Lgt: [Interval constant: 0.0]]) 
	{
	  [agent->position add: [agent->demand prod: bidfrac]];
	  [agent->cash     sub: [agent->demand prod: bfp]];
	}
      else if ([agent->demand Llt: [Interval constant: 0.0]]) 
	{
	  [agent->position add: [agent->demand prod: offerfrac]];
	  [agent->cash     sub: [agent->demand prod: ofp]];
	}
    }

  [index drop];

  [bfp drop];
  [ofp drop];
  [tp drop];
  [profitperunit drop];
  [price drop];

  return self;
}

-(void)drop {
  [maxprice drop];	
  [minprice drop];
  [eta drop];
  [minexcess drop];
  [rea drop];
  [reb drop];
  [bidfrac drop];
  [offerfrac drop];
  [volume drop];
  [taupdecay drop];
  [taupnew drop];
  [super drop];
}

@end











